@@ -43,9 +43,6 @@ static unsigned int cur_ver = 0;
 static Bit_Chain pdat = {NULL,0,0,0,0,0};
 static Bit_Chain *dat = &pdat;
 
-/* from dwg.c */
-int dwg_obj_is_control(const Dwg_Object *obj);
-
 /*--------------------------------------------------------------------------------
  * MACROS
  */
@@ -69,7 +66,10 @@ int dwg_obj_is_control(const Dwg_Object *obj);
 
 #define ANYCODE -1
 #define FIELD_HANDLE(name,code,dxf) VALUE_HANDLE(_obj->name,code,dxf)
-#define VALUE_HANDLE(hdl,code,dxf)  /* freed globally */
+#define VALUE_HANDLE(ref,code,dxf) \
+  if (ref) { \
+    if (!ref->obj && !ref->handleref.size && !ref->absolute_ref) free(ref); \
+  } /* else freed globally */
 #define FIELD_DATAHANDLE(name,code,dxf) FIELD_HANDLE(name, code, dxf)
 #define FIELD_HANDLE_N(name,vcount,code,dxf) FIELD_HANDLE(name, code, dxf)
 
@@ -93,10 +93,13 @@ int dwg_obj_is_control(const Dwg_Object *obj);
       free (FIELD_VALUE(name)); \
       FIELD_VALUE(name) = NULL; \
     }
+#define VALUE_TV(value,dxf) FREE_IF(value)
+#define VALUE_TF(value,dxf) FREE_IF(value)
+#define VALUE_TFF(value,dxf) 
 #define FIELD_TU(name,dxf)  FIELD_TV(name,dxf)
 #define FIELD_TF(name,len,dxf) FIELD_TV(name,dxf)
 #define FIELD_TFF(name,len,dxf) {}
-#define FIELD_T FIELD_TV /*TODO: implement version dependant string fields */
+#define FIELD_T(name,dxf) FIELD_TV(name,dxf)
 #define FIELD_BT(name,dxf) FIELD(name, BT);
 #define FIELD_4BITS(name,dxf) {}
 #define FIELD_BE(name,dxf) {}
@@ -111,16 +114,16 @@ int dwg_obj_is_control(const Dwg_Object *obj);
 #define FIELD_3BD_1(name,dxf) {}
 #define FIELD_3DPOINT(name,dxf) {}
 #define FIELD_TIMEBLL(name,dxf)
-#define FIELD_CMC(token, dxf)\
-  { FIELD_TV(token.name, dxf); \
-    FIELD_TV(token.book_name, dxf+1); }
+#define FIELD_CMC(color,dxf1,dxf2) \
+  { FIELD_TV(color.name, 0); \
+    FIELD_TV(color.book_name, 0); }
 
 //FIELD_VECTOR_N(name, type, size):
 // reads data of the type indicated by 'type' 'size' times and stores
 // it all in the vector called 'name'.
 #define FIELD_VECTOR_N(name, type, size, dxf) \
-  if (size) { \
-    for (vcount=0; vcount < (int)size; vcount++)  \
+  if (size && _obj->name) { \
+    for (vcount=0; vcount < (BITCODE_BL)size; vcount++)  \
       FIELD_##type(name[vcount], dxf); \
   } \
   FIELD_TV(name,dxf);
@@ -130,11 +133,13 @@ int dwg_obj_is_control(const Dwg_Object *obj);
 #define FIELD_2DD_VECTOR(name, size, dxf) FIELD_TV(name,dxf)
 #define FIELD_3DPOINT_VECTOR(name, size, dxf) FIELD_TV(name,dxf)
 #define HANDLE_VECTOR_N(name, size, code, dxf) \
-  for (vcount=0; vcount < (long)size; vcount++) \
+  if (_obj->name) { \
+    for (vcount=0; vcount < (BITCODE_BL)size; vcount++) \
     {\
       FIELD_HANDLE_N(name[vcount], vcount, code, dxf);  \
     } \
-  if (size) { FIELD_TV(name,dxf) }
+    if (size) { FIELD_TV(name,dxf) } \
+  }
 #define HANDLE_VECTOR(name, sizefield, code, dxf) \
   HANDLE_VECTOR_N(name, FIELD_VALUE(sizefield), code, dxf)
 
@@ -143,17 +148,17 @@ int dwg_obj_is_control(const Dwg_Object *obj);
   dwg_free_xdata(_obj, _obj->size)
 
 #define REACTORS(code) \
-  for (vcount=0; vcount < (long)obj->tio.object->num_reactors; vcount++) \
-    {\
+  if (obj->tio.object->reactors) { \
+    for (vcount=0; vcount < obj->tio.object->num_reactors; vcount++) \
       VALUE_HANDLE(obj->tio.object->reactors[vcount], code, 330);  \
-    } \
-    VALUE_TV(obj->tio.object->reactors, 0)
-#define ENT_REACTORS(code)  \
-  for (vcount=0; vcount < ent->num_reactors; vcount++)\
-    {\
+    VALUE_TV(obj->tio.object->reactors, 0); \
+  }
+#define ENT_REACTORS(code) \
+  if (ent->reactors) { \
+    for (vcount=0; vcount < ent->num_reactors; vcount++)\
       VALUE_HANDLE(ent->reactors[vcount], code, 330);  \
-    }\
-  VALUE_TV(ent->reactors,0)
+    VALUE_TV(ent->reactors,0); \
+  }
 #define XDICOBJHANDLE(code)\
   SINCE(R_2004)\
     {\
@@ -195,22 +200,21 @@ static int dwg_free_UNKNOWN_OBJ (Bit_Chain *restrict dat, Dwg_Object *restrict o
 
 #define DWG_ENTITY(token) \
 static int \
-dwg_free_ ##token (Bit_Chain *restrict dat, Dwg_Object *restrict obj)\
+dwg_free_ ##token (Bit_Chain *restrict _dat, Dwg_Object *restrict obj)\
 {\
-  long vcount, rcount1, rcount2, rcount3, rcount4;\
+  BITCODE_BL vcount, rcount1, rcount2, rcount3, rcount4;\
   Dwg_Entity_##token *ent, *_obj;\
   Dwg_Object_Entity *_ent;\
-  Bit_Chain *hdl_dat = dat;\
-  Bit_Chain* str_dat = dat;\
+  Bit_Chain *hdl_dat = _dat;\
+  Bit_Chain* str_dat = _dat;\
   Dwg_Data* dwg = obj->parent;\
   int error = 0; \
   LOG_HANDLE("Free entity " #token "\n")\
-  if (strcmp(#token, "UNKNOWN_ENT") && obj->supertype == DWG_SUPERTYPE_UNKNOWN) \
-    return dwg_free_UNKNOWN_ENT(dat, obj); \
   _ent = obj->tio.entity;\
   _obj = ent = _ent->tio.token;
 
 #define DWG_ENTITY_END      \
+  dwg_free_common_entity_data(obj); \
   dwg_free_eed(obj);        \
   FREE_IF(_obj);            \
   FREE_IF(obj->tio.entity); \
@@ -220,17 +224,17 @@ dwg_free_ ##token (Bit_Chain *restrict dat, Dwg_Object *restrict obj)\
 
 #define DWG_OBJECT(token) \
 static int \
-dwg_free_ ##token (Bit_Chain *restrict dat, Dwg_Object *restrict obj) \
+dwg_free_ ##token (Bit_Chain *restrict _dat, Dwg_Object *restrict obj) \
 { \
-  long vcount, rcount1, rcount2, rcount3, rcount4; \
+  BITCODE_BL vcount, rcount1, rcount2, rcount3, rcount4; \
   Dwg_Object_##token *_obj;                      \
-  Bit_Chain *hdl_dat = dat;                      \
-  Bit_Chain* str_dat = dat;                      \
+  Bit_Chain *hdl_dat = _dat;                     \
+  Bit_Chain* str_dat = _dat;                     \
   Dwg_Data* dwg = obj->parent;                   \
   int error = 0; \
   LOG_HANDLE("Free object " #token " %p\n", obj) \
   if (strcmp(#token, "UNKNOWN_OBJ") && obj->supertype == DWG_SUPERTYPE_UNKNOWN) \
-    return dwg_free_UNKNOWN_OBJ(dat, obj); \
+    return dwg_free_UNKNOWN_OBJ(_dat, obj); \
   _obj = obj->tio.object->tio.token;
 
 /* obj itself is allocated via dwg->object[], dxfname is klass->dxfname */
@@ -248,8 +252,9 @@ dwg_free_common_entity_handle_data(Dwg_Object* obj)
 
   Dwg_Data *dwg = obj->parent;
   Dwg_Object_Entity *_obj;
-  long unsigned int vcount;
+  BITCODE_BL vcount;
   Dwg_Object_Entity *ent;
+  int error = 0;
 
   ent = obj->tio.entity;
   _obj = ent;
@@ -258,6 +263,24 @@ dwg_free_common_entity_handle_data(Dwg_Object* obj)
 
 }
 
+static void
+dwg_free_common_entity_data(Dwg_Object* obj)
+{
+
+  Dwg_Data *dwg = obj->parent;
+  Dwg_Object_Entity *_obj;
+  Dwg_Object_Entity *ent;
+  Bit_Chain *hdl_dat = NULL;
+  BITCODE_BL vcount;
+  int error = 0;
+
+  ent = obj->tio.entity;
+  _obj = ent;
+
+  #include "common_entity_data.spec"
+
+}
+
 static void
 dwg_free_xdata(Dwg_Object_XRECORD *obj, int size)
 {
@@ -268,14 +291,13 @@ dwg_free_xdata(Dwg_Object_XRECORD *obj, int size)
 static void
 dwg_free_eed(Dwg_Object* obj)
 {
-  unsigned int i;
+  BITCODE_BL i;
   if (obj->supertype == DWG_SUPERTYPE_OBJECT) {
     Dwg_Object_Object* _obj = obj->tio.object;
     for (i=0; i < _obj->num_eed; i++) {
       if (_obj->eed[i].size)
         FREE_IF(_obj->eed[i].raw);
-      if (_obj->eed[i].data)
-        FREE_IF(_obj->eed[i].data);
+      FREE_IF(_obj->eed[i].data);
     }
     FREE_IF(_obj->eed);
   }
@@ -284,8 +306,7 @@ dwg_free_eed(Dwg_Object* obj)
     for (i=0; i < _obj->num_eed; i++) {
       if (_obj->eed[i].size)
         FREE_IF(_obj->eed[i].raw);
-      if (_obj->eed[i].data)
-        FREE_IF(_obj->eed[i].data);
+      FREE_IF(_obj->eed[i].data);
     }
     FREE_IF(_obj->eed);
   }
@@ -315,6 +336,9 @@ dwg_free_variable_type(Dwg_Data * dwg, Dwg_Object* obj)
   // global class dispatcher
   #include "classes.inc"
 
+  #undef WARN_UNHANDLED_CLASS
+  #undef WARN_UNSTABLE_CLASS
+  
   return DWG_ERR_UNHANDLEDCLASS;
 }
 
@@ -334,6 +358,9 @@ dwg_free_object(Dwg_Object *obj)
   if (obj->type == DWG_TYPE_FREED)
     return;
   dat->from_version = dat->version;
+  if (obj->supertype == DWG_SUPERTYPE_UNKNOWN)
+    goto unhandled;
+
   switch (obj->type)
     {
     case DWG_TYPE_TEXT:
@@ -587,8 +614,12 @@ dwg_free_object(Dwg_Object *obj)
                & DWG_ERR_UNHANDLEDCLASS)
         {
           int is_entity;
-          int i = obj->type - 500;
-          Dwg_Class *klass = NULL;
+          int i;
+          Dwg_Class *klass;
+
+        unhandled:
+          i = obj->type - 500;
+          klass = NULL;
 
           dwg = obj->parent;
           if (dwg->dwg_class && i >= 0 && i < (int)dwg->num_classes)
@@ -624,7 +655,7 @@ static int dwg_free_header_vars(Dwg_Data * dwg)
 void
 dwg_free(Dwg_Data * dwg)
 {
-  unsigned int i;
+  BITCODE_BL i;
   if (dwg)
     {
       if (dwg->opts)
@@ -663,7 +694,7 @@ dwg_free(Dwg_Data * dwg)
         }
       if (dwg->num_classes)
         {
-          for (i=0; i < (int)dwg->num_classes; ++i)
+          for (i=0; i < dwg->num_classes; ++i)
             {
               FREE_IF(dwg->dwg_class[i].appname);
               FREE_IF(dwg->dwg_class[i].cppname);
@@ -673,7 +704,7 @@ dwg_free(Dwg_Data * dwg)
             }
           FREE_IF(dwg->dwg_class);
         }
-      for (i=0; i < (int)dwg->num_object_refs; ++i)
+      for (i=0; i < dwg->num_object_refs; ++i)
         {
           FREE_IF(dwg->object_ref[i]);
         }
